/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ForecastFox.
 *
 * The Initial Developer of the Original Code is
 * Jon Stritar <jstritar@MIT.EDU>.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Jon Stritar <jstritar@MIT.EDU>
 * Richard Klein <richwklein@mchsi.com>
 * Thomas Bertels <tbertels@gmail.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
 
/* GUID */
const GUID = "{B5EDFBB0-9827-11DA-A72B-0800200C9A66}";

/* Version */
const VERSION = "0.7";
const LAST_COMPATIBLE_VERSION = "0.7";

/* Frequency Constants */
const FREQ_CURRENT = 30;
const FREQ_FORECAST = 120;
const FREQ_LINKS = 720;

/* Last Update Prefs */
const PREF_CURRENT_LAST = "cc.last";
const PREF_FORECAST_LAST = "dayf.last";
const PREF_LINKS_LAST = "links.last";

/* Cache Prefs */
const PREF_CURRENT_CACHE = "cc.cache";
const PREF_FORECAST_CACHE = "dayf.cache";
const PREF_LINKS_CACHE = "links.cache";

/* Profile Location */
const PROFILE_FILE = "profiles.xml";

/* Export Settings ID */
const EXPORT_FILE_ID = "forecastfox:settings";

/* Export File Name */
const EXPORT_FILE = "export.xml";

/* Export File "Additional Prefs" */
const EXPORT_ADDITIONAL_PREFS = [["migrated", "Char"], ["profile.current", "Char"], ["profile.switch.delay", "Int"], ["profile.switch.enabled", "Bool"]];

mfirstError = true;

/******************************************************************************
 * nsWfUpdateService class constructor
 ******************************************************************************/
function nsForecastfoxUpdate()
{
  //internal variables
  this.mProductCode = "xoap";
  this.mPartnerId   = "1005561467";
  this.mLicenseKey  = "a388d0d4a8a3f714";
  this.mServer    = "http://xoap.weather.com/weather/local/";
  this.mTimer     = null;
  this.mTimeOut   = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
  this.mHttpRequest = null;
  this.mPrefs   = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("forecastfox.");
  this.mObsService  = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
  this.mDOMParser = Components.classes["@mozilla.org/xmlextras/domparser;1"].createInstance(Components.interfaces.nsIDOMParser);
  this.mFileIO      = new fileIO();
  this.mLoadingProf = false;
  this.mFirstRun    = true;
  var sbs           = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
  this.mBundle      = sbs.createBundle("chrome://forecastfox/locale/forecastfox.properties");
  
  //set internal arrays for storage
  this.mLastError   = {};
  this.mLoc     = {};
  this.mCurrent   = {};
  this.mForecast  = new Array();
  this.mPartDay   = new Array();
  this.mPartNight = new Array();
  this.mLinks     = new Array();

  //migrate
  var migrated = this.migrate();
  
  //add shutdown observer
  this.mObsService.addObserver(this, "xpcom-shutdown", false);
  this.mObsService.addObserver(this, "forecastfox:profile:start", false);
  this.mObsService.addObserver(this, "forecastfox:profile:end", false);
  this.mObsService.addObserver(this, "forecastfox:profile:error", false);

  //add prefernce observer
  var pbi = this.mPrefs.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
  pbi.addObserver("", this, false);
}

nsForecastfoxUpdate.prototype =
{
  //readonly attributes fo UI
  get productCode()    { return this.mProductCode; },
  get partnerId()      { return this.mPartnerId;   },
  get licenseKey()     { return this.mLicenseKey;  },
  get server()         { return this.mServer;      },
  get firstRun()       { return this.mFirstRun;    },
  
  getKeys: function(aName, aCount)
  {
    var asKeys = new Array();
    var sKey = null;  
    try {
      switch (aName) {
        case "lastError":
          for (sKey in this.mLastError) asKeys.push(sKey);
          aCount.value = asKeys.length;
          break;
        case "location":
          for (sKey in this.mLoc) asKeys.push(sKey);
          aCount.value = asKeys.length;
          break;
        case "current":
          for (sKey in this.mCurrent) asKeys.push(sKey);
          aCount.value = asKeys.length;
          break;
        case "forecast":
          for (sKey in this.mForecast[0]) asKeys.push(sKey);
          aCount.value = asKeys.length;
          break;
        case "day":
          for (sKey in this.mPartDay[0]) asKeys.push(sKey);
          aCount.value = asKeys.length;
          break;
        case "night":
          for (sKey in this.mPartNight[0]) asKeys.push(sKey);
          aCount.value = asKeys.length;
          break;
        case "links":
          for (sKey in this.mLinks[0]) asKeys.push(sKey);
          aCount.value = asKeys.length;
          break;
      }
    } catch(e) {
      aCount.value = 0;
      asKeys = new Array();
    }
    
    asKeys.sort();
    return asKeys;
  },

  getLinkCount: function()
  {
    return this.mLinks.length;
  },

  getValue: function(aName, aKey, aIndex)
  {
    var val = "";
    try {
      switch (aName) {
        case "lastError":
          val = this.mLastError[aKey];
          break;
        case "location":
          val = this.mLoc[aKey];
          break;
        case "current":
          val = this.mCurrent[aKey];
          break;
        case "forecast":
          val = this.mForecast[aIndex][aKey];
          break;
        case "day":
          val = this.mPartDay[aIndex][aKey];
          break;
        case "night":
          val = this.mPartNight[aIndex][aKey];
          break;
        case "links":
          val = this.mLinks[aIndex][aKey];
          break;
      }
    } catch(e) {
      val = "";
    }
    
    return val;
  },

  load: function(aForced)
  {
    //reset last prefs if forced
    if (aForced) {
      this.mPrefs.clearUserPref(PREF_CURRENT_LAST);
      this.mPrefs.clearUserPref(PREF_FORECAST_LAST);
      this.mPrefs.clearUserPref(PREF_LINKS_LAST);
    }

    //update the data
    this.mFirstRun = false;
    this.updateData();
  },

  //main internal function to load data from server
  updateData: function()
  {
    // If the locid hasn't been set yet, prompt for it and bail out.
    if (this.mPrefs.getCharPref("general.locid") == "00000") {
      getWindow().openDialog("chrome://forecastfox/content/options.xul", "options", "chrome");
    }
    else {
      // Notify observers that updating has begun.
      this.mObsService.notifyObservers(null, "forecastfox:update:start", "");

      //clear old errors
      this.mLastError = {};

      // The array contains true if the data is being refreshed from the server.
      // If false, load the data from the cache.
      var refresh_needed      = new Array();
      refresh_needed["CURRENT"]   = this.shouldRefresh("CURRENT");
      refresh_needed["FORECAST"]  = this.shouldRefresh("FORECAST");
      refresh_needed["LINKS"]       = this.shouldRefresh("LINKS");

      // Get the cached data and the server data.
      this.getCachedData(refresh_needed);
      this.getServerData(refresh_needed, null, null);

      if (refresh_needed["CURRENT"] || refresh_needed["FORECAST"] || refresh_needed["LINKS"])
        this.mTimeOut.initWithCallback(this, 40 * 1000, Components.interfaces.nsITimer.TYPE_ONE_SHOT); // XXX fix for longer timeout

    }
  },

  getCachedData: function(refresh_needed)
  {
    // First find out what files we need to load data from.
    var file_array  = new Array();
    var failed    = new Array();
    for (var value in refresh_needed) {
      if (!refresh_needed[value]) {
        var file = this.mPrefs.getCharPref(eval("PREF_" + value + "_CACHE"));
        if (!file_array[file])
          file_array[file] = new Array();
        file_array[file][value] = true;
      }
    }

    // Then read the files.
    for (file in file_array) {
      var text  = this.mFileIO.read(file);

      if (!text) {
        for (value in file_array[file])
          if (file_array[file][value])
            failed[value] = true;
        continue;
      }

      var doc = this.mDOMParser.parseFromString(text, "text/xml");
      if (doc.childNodes[0].localName == "parsererror")
        this.getServerData(file_array[file], null, null);
      else     
        this.parseResponse(doc, file_array[file], false);
    }

    if (failed["CURRENT"] || failed["FORECAST"] || failed["LINKS"])
      this.getServerData(failed, null, null);
    else if (!refresh_needed["CURRENT"] && !refresh_needed["FORECAST"] && !refresh_needed["LINKS"])
      this.scheduleUpdate();
  },

  getServerData: function(refresh_needed, user, password)
  {
    var mComponent = this;

    // Bail out if nothing needs to be refreshed.
    if (!refresh_needed["CURRENT"] && !refresh_needed["FORECAST"] && !refresh_needed["LINKS"])
      return;

    // Construct the query url.
    var query = this.mServer + this.mPrefs.getCharPref("general.locid") + "?";

    // Only query the server for things that we need.
    if (refresh_needed["CURRENT"])
      query += "cc=*&";
    if (refresh_needed["FORECAST"])
      query += "dayf=10&";
    if (refresh_needed["LINKS"])
      query += "link=xoap&";

    // Add the unit and the license info.
    var units = this.mPrefs.getIntPref("general.units");
    
    if (units == 3)
      query += "unit=m&";
    else
      query += "unit=s&" + ((units == 1) ? "" : "ut=C&");
      
    query += "prod=" + this.mProductCode + "&par=" + this.mPartnerId + "&key=" + this.mLicenseKey;

    // Replace query with test location:
    //query = "http://web.mit.edu/jstritar/www/test.xml";
    this.mHttpRequest = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
    this.mHttpRequest.overrideMimeType("text/xml");
    this.mHttpRequest.onerror = function () {
      mComponent.handleError(null, this.status, refresh_needed);
    }
    
    this.mHttpRequest.onload = function () {
      // It didn't timeout, so cancel that timer.
      mComponent.mTimeOut.cancel();

      var doc  = this.responseXML;

      if (mComponent.handleError(doc, this.status, refresh_needed))
        return;

      // Cache the response, then parse it and update our data.
      mComponent.saveResponse(this.responseText, refresh_needed);
      mComponent.parseResponse(doc, refresh_needed, true);

      // Then schedule an update.
      mComponent.scheduleUpdate();

      return;
    }
    if (!user || !password)
      this.mHttpRequest.open("GET", query, true);
    else
      this.mHttpRequest.open("GET", query, true, user, password);
    // Should insert some error handling for when the send timeouts.
    this.mHttpRequest.send(null);

  },

  handleError: function(doc, status, updated)
  {
    //need to authenticate
    if (status == 407) {
      //XXX prompt usernameandpassword then pass to getserver data  
    }
    
    if (status > 200) {
      this.mLastError["type"] = "0";
      this.mLastError["msg"] = this.mBundle.GetStringFromName("ff.error.net.label");
      this.mLastError["tip"] = this.mBundle.formatStringFromName("ff.error.net.tooltip", [status]);

      // Then notify the observers of the error.
      this.notifyError(updated);
      return true;
    }

    if (doc.getElementsByTagName("error").length > 0) {
      var error = doc.getElementsByTagName("err")[0];

      // Record error.
      this.mLastError["type"] = error.getAttribute("type");
      this.mLastError["msg"]  = error.firstChild.nodeValue;
      this.mLastError["tip"] = error.firstChild.nodeValue;
      
      //xxx localization of the error messages
      
    try
    {
      this.mLastError["msg"] = this.mBundle.GetStringFromName("ff.error." + this.mLastError["type"]);
      this.mLastError["tip"] = this.mBundle.GetStringFromName("ff.error." + this.mLastError["type"]);
    }
    catch(e){}

      // end of localization of the error messages
      
      // Then notify the observers of the error.
      this.notifyError(updated);

      return true;
    }
    else if (doc.childNodes[0].localName == "parsererror") {
      this.mLastError["type"] = "0";
      this.mLastError["msg"] = this.mBundle.GetStringFromName("ff.error.parse.label");
      this.mLastError["tip"] = this.mBundle.GetStringFromName("ff.error.parse.tooltip");
      this.scheduleUpdate(10 * 1000); // XXX fix for retry when parse error

      // Then notify the observers of the error.
      this.notifyError(updated);
      return true;      
    }
    
    return false;
  },
  
  notifyError: function(updated)
  {
    for (var type in updated)
      if (updated[type])
        this.mObsService.notifyObservers(null, "forecastfox:update:error", type);
  },
  
  saveResponse: function(text, refresh_needed)
  {
    var file    = null;
    var time    = (new Date()).valueOf();
    var profile = this.mPrefs.getCharPref("profile.current");

    // Assign file names this way so that the links one gets held longest.
    if (refresh_needed["LINKS"])
      file = "linkscache-" + profile + ".xml";
    else if (refresh_needed["FORECAST"])
      file = "forecastcache-" + profile + ".xml";
    else if (refresh_needed["CURRENT"])
      file = "currentcache-" + profile + ".xml";

    // Then save the file...
    this.mFileIO.write(file, text);

    // And update the CACHE prefs.
    for (var value in refresh_needed) {
      if (refresh_needed[value]) {
        this.mPrefs.setCharPref(eval("PREF_" + value + "_CACHE"), file);
        this.mPrefs.setCharPref(eval("PREF_" + value + "_LAST"), time);
      }
    }
  },

  parseResponse: function(doc, contains, server)
  {
    //XXX should have a better way of parsing this without having to know the structure of the XML
    var w,x,y,z;
    var time      = (new Date()).valueOf();

    // TODO: clear this.mLoc
    // Parse loc info.
    this.mLoc = {};
    var data_needed = new Array("locale", "ut", "ud", "us", "up", "ur", "dnam", "tm", "lat", "lon", "sunr", "suns", "zone");

    for (var value in data_needed)
      this.mLoc[data_needed[value]] = doc.getElementsByTagName(data_needed[value])[0].firstChild.nodeValue;

    this.addUnits("loc");

    if (contains["CURRENT"]) {
      // Clear old data:
      this.mCurrent = {};

      // Then parse in the new data:
      var cc = doc.getElementsByTagName("cc")[0].childNodes;

      for (x = 0; x < cc.length; x++) {
        if (cc[x].nodeName == "#text")
          continue;
        if (cc[x].nodeName == "bar" || cc[x].nodeName == "wind" || 
            cc[x].nodeName == "uv" || cc[x].nodeName == "moon") {
          var next_nodes = cc[x].childNodes;
          for (w=0; w < next_nodes.length; w++) {
            if (next_nodes[w].nodeName == "#text")
              continue;
            this.mCurrent[cc[x].nodeName + next_nodes[w].nodeName] = next_nodes[w].firstChild.nodeValue;
          }
        }
        else
          this.mCurrent[cc[x].nodeName] = cc[x].firstChild.nodeValue;
      }

      this.addUnits("current");
      this.mObsService.notifyObservers(null, "forecastfox:update:end:current", "");

      //do alert notification
      try {
        var sliderfreq = this.mPrefs.getIntPref("cc.slider.freq");
        var slidercount = this.mPrefs.getIntPref("cc.slider.count");
        var sliderenabled = this.mPrefs.getBoolPref("cc.slider.enabled");
        if (sliderenabled && server) {
          this.mPrefs.setIntPref("cc.slider.count", (slidercount + 1));
          if (slidercount >= sliderfreq) {
            // Show the slider and reset it's count.
            var alerts, alertTitle, alertText, alertURL;
            var alertDisplay = this.mPrefs.getIntPref("cc.slider.display");
            switch (alertDisplay) {
              case 0:
                alertURL = "chrome://forecastfox/skin/images/large/" + this.mCurrent["icon"] + ".png";
                alertTitle = "";
                alertText = "";
                break;
              case 1:
                alertURL = "";
                alertTitle = this.setLabel("cc.slider.title");
                alertText = this.setLabel("cc.slider.label");
                break;
              case 2:
                alertURL = "chrome://forecastfox/skin/images/large/" + this.mCurrent["icon"] + ".png";
                alertTitle = this.setLabel("cc.slider.title");
                alertText = this.setLabel("cc.slider.label");
                break;                        
            }
            
            alerts = Components.classes["@mozilla.org/alerts-service;1"].getService(Components.interfaces.nsIAlertsService);
            alerts.showAlertNotification(alertURL, alertTitle, alertText, true, "", this);
            this.mPrefs.setIntPref("cc.slider.count", "1");
          }
        }
      } catch(e) {}
    }

    if (contains["FORECAST"]) {
      // Clear old forecast data:
      this.mForecast = new Array();
      this.mPartDay = new Array();
      this.mPartNight = new Array();

      
      //get all the days in the forecast

      var days = doc.getElementsByTagName("day");
      for (x=0; x < days.length; x++) {
        var data = {};
        data["lsup"] = doc.getElementsByTagName("dayf")[0].getElementsByTagName("lsup")[0].firstChild.nodeValue;
        var day_nodes = days[x].childNodes;
        next_nodes = null;

        data["day"] = days[x].getAttribute("t");
        data["date"] = days[x].getAttribute("dt");

        //loop through nodes in day
        for (y=0; y < day_nodes.length; y++) {
          if (day_nodes[y].nodeName == "#text")
            continue;
          //get part data
          if (day_nodes[y].nodeName == "part") {
            var part_nodes = day_nodes[y].childNodes;
            var part = {};
            if (day_nodes[y].getAttribute("p") == "d") {
              //loop through nodes in part
              for (z=0; z < part_nodes.length; z++) {
                if (part_nodes[z].nodeName == "#text")
                  continue;
                if (part_nodes[z].nodeName == "bar" || part_nodes[z].nodeName == "wind" || part_nodes[z].nodeName == "uv") {
                  next_nodes = part_nodes[z].childNodes;
                  for (w=0; w < next_nodes.length; w++) {
                    if (next_nodes[w].nodeName == "#text")
                      continue;
                    part[part_nodes[z].nodeName + next_nodes[w].nodeName] = next_nodes[w].firstChild.nodeValue;
                  }
                }
                else
                  part[part_nodes[z].nodeName] = part_nodes[z].firstChild.nodeValue;
              }
              this.mPartDay[x] = part;
            }
            else {
              //loop through nodes in part
              for (z=0; z < part_nodes.length; z++) {
                if (part_nodes[z].nodeName == "#text")
                  continue;
                if (part_nodes[z].nodeName == "bar" || part_nodes[z].nodeName == "wind" || part_nodes[z].nodeName == "uv") {
                  next_nodes = part_nodes[z].childNodes;
                  for (w=0; w < next_nodes.length; w++) {
                    if (next_nodes[w].nodeName == "#text")
                      continue;
                    part[part_nodes[z].nodeName + next_nodes[w].nodeName] = next_nodes[w].firstChild.nodeValue;
                  }
                }
                else
                  part[part_nodes[z].nodeName] = part_nodes[z].firstChild.nodeValue;
              }
              this.mPartNight[x] = part;
            }
          }
          else {
            if (day_nodes[y].nodeName == "bar" || day_nodes[y].nodeName == "wind" || day_nodes[y].nodeName == "uv") {
              next_nodes = day_nodes[y].childNodes;
              for (w=0; w < next_nodes.length; w++) {
                if (next_nodes[w].nodeName == "#text")
                  continue;
                data[day_nodes[y].nodeName + next_nodes[w].nodeName] = next_nodes[w].firstChild.nodeValue;
              }
            }
            else
              data[day_nodes[y].nodeName] = day_nodes[y].firstChild.nodeValue;
          }
        }
        this.mForecast[x] = data;
      }

      this.addUnits("forecast");
      this.mObsService.notifyObservers(null, "forecastfox:update:end:forecast", "");
    }

    if (contains["LINKS"]) {
      // Parse links info.
      this.mLinks = new Array();

      var links = doc.getElementsByTagName("lnks");

      if (links[0].getAttribute("type") == "prmo")
        links = links[0].getElementsByTagName("link");

      for (x = 0; x < links.length; x++) {
        // Create the array for this link
        var link_array    = {};
        
        link_array["pos"]   = links[x].getAttribute("pos");
        link_array["url"]   = links[x].getElementsByTagName("l")[0].firstChild.nodeValue;
        link_array["title"] = links[x].getElementsByTagName("t")[0].firstChild.nodeValue;

        // Then insert it into our internal array that holds the links.
        this.mLinks[x] = link_array;
      }
      this.addUnits("links");
      this.mObsService.notifyObservers(null, "forecastfox:update:end:links", "");
    }
  },

  // xxx start of main part of localization  


  getlocalet: function(temps)
	{
    temps = temps.toLowerCase();
    temps = temps.replace("/", "");
    for (var x = 0; x < 11; x++)
      temps = temps.replace(" ", "");
    var oldtemps = temps;
    try
    {
      temps = this.mBundle.GetStringFromName("ff.t." + temps);
    }
    catch(e){}
    
    if (temps == oldtemps)
    {
      temps = "";
    }
    return temps;
	},

  setLabel: function(aPref)
  { 
    var text;
    if (this.mPrefs.prefHasUserValue(aPref))
      text = this.mPrefs.getComplexValue(aPref,Components.interfaces.nsISupportsString).data;
    else
      text = this.mBundle.GetStringFromName(this.mPrefs.getCharPref(aPref));
      
    var lieu = this.mLoc['dnam'];
    
    var country = lieu.substr(lieu.lastIndexOf(",")+2);
    var city = lieu.substr(0, lieu.indexOf(","));
    var countryorig = country;
    var cityorig = city;
    country = country.replace(" ", "");
    country = country.replace(" ", "");
    country = country.replace(" ", "");
    city = city.replace(" ", "");
    city = city.replace(" ", "");
    city = city.replace(" ", "");
    
    if  (locale == "zh-TW")
      {
		lieu = countryorig + " " + cityorig;
      }
    
    try{
      lieu = lieu.replace(countryorig, this.mBundle.GetStringFromName("ff.dnam." + country.toLowerCase()));
    }
    catch(e) {}
    try{
      lieu = lieu.replace(cityorig, this.mBundle.GetStringFromName("ff.dnam." + city.toLowerCase()));
    }
    catch(e) {}
    
    this.mLoc['dnam'] = lieu;
    
    // locale t
    
    var temps = this.mCurrent['t'];
    
    temps = this.getlocalet(temps);
    try{
    var locale = this.mBundle.GetStringFromName("ff.locale");
    }
    catch(e) {}
    
    if (temps == "")
    {
      temps = this.mCurrent['t'];
      if (locale == "fr-FR")
      {
        temps = temps.replace(" Early", "");
        temps = temps.replace(" Late", "");
        
        temps = this.getlocalet(temps);
      }
      else
        if (locale == "nl-NL")
        {
          if (temps.indexOf(" Early") != -1)
          {
            var foundtemps = " Early";
            var newtemps = "Vroeg ";
          }
          else if (temps.indexOf(" Late") != -1)
          {
            var foundtemps = " Late";
            var newtemps = "Later ";
          }
          if (foundtemps != "undefined")
          {
            temps = temps.replace(foundtemps, "");
            temps = this.getlocalet(temps);
            
            if (temps.indexOf("IJ") == 0)
              temps = newtemps + "ij" + temps.replace("IJ", "");
            else
              temps = newtemps + temps.charAt(0).toLowerCase() + temps.substring(1);
          }
        }
        else
          temps = this.mCurrent['t'];
          
      if (locale != "undefined")
      {
        if (temps == "")
          temps = this.mCurrent['t'];
      }
    }
  
  // end of main part of localization  

    var base = [[/\[locale\]/g, "this.mLoc['locale']"],
                [/\[prof\]/g, "this.mPrefs.getCharPref('profile.current')"],
                [/\[ut\]/g, "this.mLoc['ut']"],
                [/\[ud\]/g, "this.mLoc['ud']"],
                [/\[us\]/g, "this.mLoc['us']"],
                [/\[up\]/g, "this.mLoc['up']"],
                [/\[ur\]/g, "this.mLoc['ur']"],
                [/\[dnam\]/g, "this.mLoc['dnam']"],
                [/\[tm\]/g, "this.mLoc['tm']"],
                [/\[lat\]/g, "this.mLoc['lat']"],
                [/\[lon\]/g, "this.mLoc['lon']"],
                [/\[sunr\]/g, "this.mLoc['sunr']"],
                [/\[suns\]/g, "this.mLoc['suns']"],
                [/\[zone\]/g, "this.mLoc['zone']"],
                [/\[deg\]/g, "String.fromCharCode(176)"],
                [/\[lsup\]/g, "this.mCurrent['lsup']"],
                [/\[obst\]/g, "this.mCurrent['obst']"],
                [/\[tmp\]/g, "this.mCurrent['tmp']"],
                [/\[flik\]/g, "this.mCurrent['flik']"],
                [/\[t\]/g, "temps"],
                [/\[icon\]/g, "this.mCurrent['icon']"],
                [/\[barr\]/g, "this.mCurrent['barr']"],
                [/\[bard\]/g, "this.mCurrent['bard']"],
                [/\[winds\]/g, "this.mCurrent['winds']"],
                [/\[windgust\]/g, "this.mCurrent['windgust']"],
                [/\[windd\]/g, "this.mCurrent['windd']"],
                [/\[windt\]/g, "this.mCurrent['windt']"],
                [/\[hmid\]/g, "this.mCurrent['hmid']"],
                [/\[vis\]/g, "this.mCurrent['vis']"],
                [/\[uvi\]/g, "this.mCurrent['uvi']"],
                [/\[uvt\]/g, "this.mCurrent['uvt']"],
                [/\[moont\]/g, "this.mCurrent['moont']"],
                [/\[dewp\]/g, "this.mCurrent['dewp']"]];

    // Then replace each one with the value.
    for (var x = 0; x < base.length; x++)
      text = text.replace(base[x][0], eval(base[x][1]));
    return text;    
  },
  
  addUnits: function(type)
  {
    var deg = String.fromCharCode(176);

    switch (type) {
      case "loc" :
        this.mLoc["lat"]          = addUnit(this.mLoc["lat"], deg);
        this.mLoc["lon"]          = addUnit(this.mLoc["lon"], deg);
        break;
      case "current" :
        this.mCurrent["tmp"]    = addUnit(this.mCurrent["tmp"], deg + this.mLoc["ut"]);
        this.mCurrent["flik"]     = addUnit(this.mCurrent["flik"], deg + this.mLoc["ut"]);
        this.mCurrent["barr"]     = addUnit(this.mCurrent["barr"], this.mLoc["ur"], " ");
        this.mCurrent["winds"]    = addUnit(this.mCurrent["winds"], this.mLoc["us"], " ");
        this.mCurrent["windgust"] = addUnit(this.mCurrent["windgust"], this.mLoc["us"], " ");
        this.mCurrent["windd"]    = addUnit(this.mCurrent["windd"], deg);
        this.mCurrent["hmid"]     = addUnit(this.mCurrent["hmid"], "%");
        this.mCurrent["vis"]    = addUnit(this.mCurrent["vis"], this.mLoc["ud"], " ");
        this.mCurrent["dewp"]     = addUnit(this.mCurrent["dewp"], deg + this.mLoc["ut"]);
        break;
      case "forecast" :
        for (var x = 0; x < this.mForecast.length; x++) {
          this.mForecast[x]["hi"]      = addUnit(this.mForecast[x]["hi"], deg + this.mLoc["ut"]);
          this.mForecast[x]["low"]     = addUnit(this.mForecast[x]["low"], deg + this.mLoc["ut"]);
          this.mPartDay[x]["winds"]    = addUnit(this.mPartDay[x]["winds"], this.mLoc["us"], " ");
          this.mPartDay[x]["windgust"]   = addUnit(this.mPartDay[x]["windgust"], this.mLoc["us"], " ");
          this.mPartDay[x]["windd"]    = addUnit(this.mPartDay[x]["daywindd"], deg);
          this.mPartDay[x]["ppcp"]     = addUnit(this.mPartDay[x]["ppcp"], "%");
          this.mPartDay[x]["hmid"]     = addUnit(this.mPartDay[x]["hmid"], "%");
          this.mPartNight[x]["winds"]    = addUnit(this.mPartNight[x]["winds"], this.mLoc["us"], " ");
          this.mPartNight[x]["windgust"] = addUnit(this.mPartNight[x]["windgust"], this.mLoc["us"], " ");
          this.mPartNight[x]["windd"]    = addUnit(this.mPartNight[x]["windd"], deg);
          this.mPartNight[x]["ppcp"]   = addUnit(this.mPartNight[x]["ppcp"], "%");
          this.mPartNight[x]["hmid"]   = addUnit(this.mPartNight[x]["hmid"], "%");
        }
        break;
      case "links" :
        break;
    }
  },

  migrate: function()
  {
    switch (this.mPrefs.getCharPref("migrated")) {
      /*case VERSION+"*":
        // completes the uninstallation/upgrade of weatherfox
        var wPrefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("weatherfox.");
        var dirService  = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);

        // remove all weatherfox. preferences   
        wPrefs.deleteBranch("");
      
        // now get rid of the weatherfox folder in the profile directory
        var wFolder = dirService.get("ProfD", Components.interfaces.nsIFile);
        wFolder.append("weatherfox");
      
        if (wFolder.exists())
          wFolder.remove(true);
             
        break;
        */
      case VERSION:
        // already migrated - return that we're ready to load
        return false;
       /*
      default:
        // migrate from older versions that do not use the current migrate pref
        
        var prefList = [["locid", "general.locid", "Char"], ["celcius", "general.units", "Bool"],
                       ["display.bar", "general.bar", "Char"], ["current.display", "cc.panel.enabled", "Bool"],
                       ["current.mode", "cc.panel.display", "Int"], ["current.tooltip", "cc.tooltip.label", "Char"],
                       ["current.label", "cc.panel.label", "Char"], ["forecast.days", "dayf.panel.days", "Int"],
                       ["forecast.mode", "dayf.panel.display", "Int"], ["forecast.parts", "dayf.panel.mode", "Int"],
                       ["forecast.label", "dayf.panel.label", "Char"], ["forecast.tooltip", "dayf.tooltip.label", "Char"],
                       ["slider.freq", "cc.slider.freq", "Int"], ["profile", "profile.current", "Char"],
                       ["profile.switch", "profile.switch.delay", "Int"]];
        //get the extension manager
        try {
          var em = Components.classes["@mozilla.org/extensions/manager;1"].getService(Components.interfaces.nsIExtensionManager);
          var nsIUpdateItem = Components.interfaces.nsIUpdateItem;
        } catch (e) {
          // if there is no em, then we're installing on Mozilla... therefore no migration is needed.
          return false;
        }
        
        //if weatherfox is still listed then uninstall
        var wfox = em.getItemList("{BA8E053E-2867-4772-B9F0-26A5979AFA09}", nsIUpdateItem.TYPE_EXTENSION, {});
        var isWfoxUdating = (wfox[0] != null);
        if (isWfoxUdating)
          // migrate from Weatherfox... convert prefs that can be converted
          var wPrefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("weatherfox.");
        var prefService = isWfoxUdating ? wPrefs : this.mPrefs;
        
        for (var x = 0; x < prefList.length; x++) {
          try {
            if (prefService.prefHasUserValue(prefList[x][0]))
              eval("this.mPrefs.set" + prefList[x][2] + "Pref('" + prefList[x][1] + "', prefService.get" + 
                prefList[x][2] + "Pref('" + prefList[x][0] + "'))");
            if (!isWfoxUdating)
              this.mPrefs.clearUserPref(prefList[x][0]);
          } catch (e) {}
        }
          
        this.mPrefs.setIntPref("dayt.panel.display", this.mPrefs.getIntPref("dayf.panel.display"));
        this.mPrefs.setIntPref("dayt.panel.mode", this.mPrefs.getIntPref("dayf.panel.mode"));
        var days = this.mPrefs.getIntPref("dayf.panel.days")
        
        if (days > 1)
          this.mPrefs.setIntPref("dayf.panel.days", days - 1);
        else if (days == 1)
          this.mPrefs.setBoolPref("dayf.panel.enabled", false);
        else {
          this.mPrefs.setBoolPref("dayf.panel.enabled", false);
          this.mPrefs.setBoolPref("dayt.panel.enabled", false);
        }
        
        if (this.mPrefs.prefHasUserValue("dayf.panel.label"))
          this.mPrefs.setCharPref("dayt.panel.label", this.mPrefs.getCharPref("dayf.panel.label"));
        if (this.mPrefs.prefHasUserValue("dayf.tooltip.label"))
          this.mPrefs.setCharPref("dayt.tooltip.label", this.mPrefs.getCharPref("dayf.tooltip.label"));
        
        // now finish the migrating up by...
        if (isWfoxUdating) {
          // .. unintsalling weatherfox
          em.uninstallExtension("{BA8E053E-2867-4772-B9F0-26A5979AFA09}");     
          
          //prompt restart
          var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
          var bundle = sbs.createBundle("chrome://forecastfox/locale/forecastfox.properties");
          var ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);    
          ps.alert(getWindow(), bundle.GetStringFromName("ff.restart.title"), bundle.GetStringFromName("ff.restart.message"));
          
          // add * to complete the migration the next time Firefox starts
          this.mPrefs.setCharPref("migrated", VERSION+"*");
          
          return true;
        }
        else {
          // ... or clearing the old forecastfox prefs
          var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("forecastfox.current.");
          prefs.deleteBranch("");
          prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("forecastfox.forecast.");
          prefs.deleteBranch("");
          prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("forecastfox.slider.");
          prefs.deleteBranch("");
          prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("forecastfox.display.");
          prefs.deleteBranch("");
        }*/
        break;
    }
    

    this.mPrefs.setCharPref("migrated", VERSION);
    
    //return that it is okay to load
    return false;
  },

  scheduleUpdate: function(force_time)
  {
    var timer_time = null;

    if (force_time)
      timer_time = force_time;
    else {

      // Schedules the next updateData call.
      var real_time = (new Date()).valueOf();

      var next    = new Array();
      // Get time till next refresh for each category.
      next[0]   = Number(this.mPrefs.getCharPref(PREF_CURRENT_LAST)) + FREQ_CURRENT * 60 * 1000 - real_time;
      next[1]     = Number(this.mPrefs.getCharPref(PREF_FORECAST_LAST)) + FREQ_FORECAST * 60 * 1000 - real_time;
      next[2]   = Number(this.mPrefs.getCharPref(PREF_LINKS_LAST)) + FREQ_LINKS * 60 * 1000 - real_time;

      next.sort(function (a, b) { return (a - b) });

      // If there will be another refresh within 1 min of the scheduled one, refresh
      // at that point instead --> to clump them together.

      if ((next[2] - next[0]) < 15 * 60 * 1000)
        timer_time = next[2];
      else if ((next[1] - next[0]) < 10 * 60 * 1000)
        timer_time = next[1];
      else
        timer_time = next[0];

      if (timer_time < 0)
        timer_time = 100;

      timer_time += 100;

    }

    // Create the timer if its not already made
    if (!this.mTimer)
      this.mTimer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);

    this.mTimer.cancel();

    // Then schedule the timeout.
    this.mTimer.initWithCallback(this, timer_time, Components.interfaces.nsITimer.TYPE_ONE_SHOT);
  },

  shouldRefresh: function(data)
  {
    var time = (new Date()).valueOf();

    // Make sure data is a valid element.
    if ((data != "CURRENT") && (data != "FORECAST") && (data != "LINKS"))
      return false;

    // Then get the time for the next refresh.
    var next = Number(this.mPrefs.getCharPref(eval("PREF_" + data + "_LAST"))) + eval("FREQ_" + data) * 60 * 1000;

    return ((next - time) <= 0);
  },

  ///////////////////////////
  // nsITimerCallback
  notify: function(timer)
  {
    if (timer == this.mTimeOut) {
      if (mfirstError) { // XXX fix for quick first timeout retry
        this.scheduleUpdate(10 * 1000);
        mfirstError = false;
      }
      else
        this.scheduleUpdate(2 * 60 * 1000);

      this.mLastError["type"] = "0";
      this.mLastError["msg"] = this.mBundle.GetStringFromName("ff.error.timeout.label");
      this.mLastError["tip"] = this.mBundle.GetStringFromName("ff.error.timeout.tooltip");

      //XXX I don't like having to recalculate these... other than using something like this.refresh_needed
      //    i'm not sure how to do it
      
      var refresh_needed            = new Array();
      refresh_needed["CURRENT"]   = this.shouldRefresh("CURRENT");
      refresh_needed["FORECAST"]  = this.shouldRefresh("FORECAST");
      refresh_needed["LINKS"]       = this.shouldRefresh("LINKS");       
 
     // Then notify the observers of the error.
      this.notifyError(refresh_needed);
    }
    else if (timer == this.mTimer) {
      // update the data.
      this.updateData();
    }
  },

  ///////////////////////////
  // nsIAlertListener
  onAlertFinished: function() {},
  onAlertClickCallback: function(aCookie)
  {
    var browser = getWindow().document.getElementById("content");
    try{
		var locale = this.mBundle.GetStringFromName("ff.locale");
		}
		catch(e) {}
	if (locale == "fr-FR")
	    var url = "http://fr.weather.com/weather/local/" + this.mPrefs.getCharPref("general.locid"); //FR
	else
		var url = "http://www.weather.com/weather/local/" + this.mPrefs.getCharPref("general.locid");
    url = url + "?prod=" + this.mProductCode + "&par=" + this.mPartnerId;
    browser.loadURI(url);
  },

  /////////////////////////////
  // nsIObserver
  observe: function(aSubject, aTopic, aData)
  {
    switch (aTopic) {
      case "forecastfox:profile:start":
        this.mLoadingProf = true;
        break;
      case "forecastfox:profile:end":
        this.mLoadingProf = false;
        this.scheduleUpdate(100);
        break;
      case "forecastfox:profile:error":
        this.mLoadingProf = false;
        break;
      case "nsPref:changed":
        if (!this.mLoadingProf) {
          if (aData == "general.locid" || aData == "general.units") {
            if (this.mPrefs.prefHasUserValue(PREF_CURRENT_LAST))
              this.mPrefs.clearUserPref(PREF_CURRENT_LAST);
            if (this.mPrefs.prefHasUserValue(PREF_FORECAST_LAST))
              this.mPrefs.clearUserPref(PREF_FORECAST_LAST);
            if (this.mPrefs.prefHasUserValue(PREF_LINKS_LAST))
              this.mPrefs.clearUserPref(PREF_LINKS_LAST);
            this.scheduleUpdate(250);
          }
        }
        break;
      case "xpcom-shutdown":

        // Clean up held observers etc to avoid leaks.
        this.mObsService.removeObserver(this, "xpcom-shutdown");
        this.mObsService.removeObserver(this, "forecastfox:profile:start", false);
        this.mObsService.removeObserver(this, "forecastfox:profile:end", false);
        this.mObsService.removeObserver(this, "forecastfox:profile:error", false);
        var pbi = this.mPrefs.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
        pbi.removeObserver("", this);

        // Release strongly held services.
        this.mPrefs        = null;
        this.mObsService   = null;
        this.mHttpRequest  = null;
        this.mDOMParser    = null;
        this.mBundle       = null;
        
        this.mFileIO.destroy();
        this.mFileIO       = null;
        if (this.mTimer) {
          this.mTimer.cancel();
          this.mTimer = null;
        }
        break;
    }
  },

  QueryInterface: function (iid)
  {
    if (!iid.equals(Components.interfaces.nsIForecastfoxUpdate) &&
          !iid.equals(Components.interfaces.nsIAlertListener)
          && !iid.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
}

function nsForecastfoxProfiles()
{
  //set up base variables
  this.mPrefs   = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("forecastfox.");
  this.mObsService  = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
  this.mDOMParser = Components.classes["@mozilla.org/xmlextras/domparser;1"].createInstance(Components.interfaces.nsIDOMParser);
  this.mSerializer  = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].createInstance(Components.interfaces.nsIDOMSerializer);
  this.mTimer       = null;
  this.mLastProfile = null;
  this.mFileIO      = new fileIO();
  this.mDirService  = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);

  this.mPrefList = [["general.locid", "Char"],
                    ["general.units", "Int"],
                    ["general.bar", "Char"],
                    ["general.position", "Int"],
                    ["cc.last", "Char"],
                    ["cc.cache", "Char"],                    
                    ["cc.panel.enabled", "Bool"],
                    ["cc.panel.display", "Int"],
                    ["cc.panel.label", "Char"],                    
                    ["cc.tooltip.enabled", "Bool"],
                    ["cc.tooltip.display", "Int"],
                    ["cc.tooltip.label", "Char"],
                    ["cc.slider.freq", "Int"],
                    ["cc.slider.count", "Int"],                                                                   
                    ["cc.slider.enabled", "Bool"], 
                    ["cc.slider.display", "Int"],
                    ["cc.slider.title", "Char"],                     
                    ["cc.slider.label", "Char"],                                                           
                    ["dayt.panel.mode", "Int"],                
                    ["dayt.panel.enabled", "Bool"],
                    ["dayt.panel.display", "Int"],
                    ["dayt.panel.label", "Char"],
                    ["dayt.panel.switch", "Int"],                  
                    ["dayt.tooltip.enabled", "Bool"],
                    ["dayt.tooltip.display", "Int"],
                    ["dayt.tooltip.label", "Char"],                    
                    ["dayf.last", "Char"],
                    ["dayf.cache", "Char"],
                    ["dayf.panel.mode", "Int"],
                    ["dayf.panel.days", "Int"],                    
                    ["dayf.panel.enabled", "Bool"],
                    ["dayf.panel.display", "Int"],
                    ["dayf.panel.label", "Char"],                    
                    ["dayf.tooltip.enabled", "Bool"],
                    ["dayf.tooltip.display", "Int"],
                    ["dayf.tooltip.label", "Char"],
                    ["links.last", "Char"],
                    ["links.cache", "Char"]];
                                       
  this.mImporting = false;
  this.mDoc = this.getProfiles();
  this.mProfile = this.mPrefs.getCharPref("profile.current");
  
  //add shutdown observer
  this.mObsService.addObserver(this, "xpcom-shutdown", false);

  //add prefernce observer for locid
  var pbi = this.mPrefs.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
  pbi.addObserver("", this, false);

  // start rotation
  this.rotateStop = false;
  this.scheduleRotate();
  return this;
}

nsForecastfoxProfiles.prototype =
{
  saveCurrent: function()
  {
    this.saveProfile(this.mProfile);
  },

  saveProfile: function(name)
  {
    name = this.stripName(name);
    var profile = this.getProfileNode(name);
    var profiles = this.mDoc.getElementsByTagName("profiles")[0];
    var new_profile = false;
    var pref;
    
    if (profile)
      profile.parentNode.removeChild(profile);
    else {
      // set new_profile so we can rename the cache files down there after the profile node has been created
      new_profile = true;
      this.mObsService.notifyObservers(null, "forecastfox:profile:new", name);
    }

    profile = this.mDoc.createElement("profile");
    profile.setAttribute("id", name);

    for (var x = 0; x < this.mPrefList.length; x++) {
      pref = this.mDoc.createElement("pref");
      pref.setAttribute("name", this.mPrefList[x][0]);
      pref.setAttribute("type", this.mPrefList[x][1]);
      pref.setAttribute("value", eval("this.mPrefs.get" + this.mPrefList[x][1] + "Pref('" + this.mPrefList[x][0] + "')"));
      profile.appendChild(pref);
    }
    
    profiles.appendChild(profile);
    
    if (new_profile)
      this.copyCacheFiles(profile.getAttribute("id"), profile);

    this.saveProfiles();
  },
  
  stopRotation: function()
  {
    this.rotateStop = true;
    
    if (!this.mTimer)
      return
    else
      this.mTimer.cancel();
  },
  
  startRotation: function()
  {
    this.rotateStop = false;
    this.scheduleRotate();
  },
  
  scheduleRotate: function()
  {
    if (this.rotateStop || this.getProfiles().getElementsByTagName("profile").length == 1)
      return;

    var rotate_time = this.mPrefs.getIntPref("profile.switch.delay");
    var rotate      = this.mPrefs.getBoolPref("profile.switch.enabled");
    
    // create timer if not already created
    if (!this.mTimer)
      this.mTimer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);

    // clear old timer
    this.mTimer.cancel();
    
    // bail out if its disabled
    if ((rotate == false) || (rotate_time == 0))
      return;
    else
      rotate_time *= 1000 * 60;

    // then schedule the timer
    this.mTimer.initWithCallback(this, rotate_time, Components.interfaces.nsITimer.TYPE_ONE_SHOT);
  },
  
  rotateProfile: function()
  {
    var profiles = this.getProfiles().getElementsByTagName("profile");
    
    // go to next profile in the list
    for (var x = 0; x < profiles.length; x++) {
      if (profiles[x].getAttribute("id") == this.mProfile) {
        this.mPrefs.setCharPref("profile.current", profiles[( x + 1 < profiles.length ) ? x + 1 : 0 ].getAttribute("id"));
        return;
      }
    }
  },
  
  getProfilePref: function(profile, pref)
  {
    var profile_node = this.getProfileNode(profile);
    var prefs = profile_node.getElementsByTagName("pref");
    for (var x = 0; x < prefs.length; x++) {
      if (prefs[x].getAttribute("name") == pref) {
        return prefs[x].getAttribute("value");
      }
    }
    return false;
  },
  
  setProfilePref: function(profile, pref, value)
  {
    var prefs = profile.getElementsByTagName("pref");
    var pref_node = null;
    
    for (var x = 0; x < prefs.length; x++) {
      if (prefs[x].getAttribute("name") == pref) {
        prefs[x].setAttribute("value", value);
        return;
      }
    }
  },
  
  getNames: function(aCount)
  {
    var asKeys = new Array();
    var sKey   = null;
    var profiles = this.mDoc.getElementsByTagName("profile");

    if (profiles.length == 0) {
      this.saveCurrent();
      profiles = this.mDoc.getElementsByTagName("profile");
    }

    for (var x = 0; x < profiles.length; x++)
      asKeys.push(profiles[x].getAttribute("id"));

    aCount.value = asKeys.length;
    asKeys.sort();
    return asKeys;
  },
  
  getKeys: function(aProfile, aCount)
  {
    var asKeys = new Array();
    
    for (var x = 0; x < this.mPrefList.length; x++) {
      asKeys.push(this.mPrefList[x][0]);
    }
    aCount.value = asKeys.length;
    
    asKeys.sort();
    return asKeys;
  },
  
  getValue: function(aProfile, aKey)
  {
    return this.getProfilePref(aProfile, aKey);
  },

  getProfileNode: function(name)
  {
    var profiles = this.mDoc.getElementsByTagName("profile");
    var val;
    for (var x = 0; x < profiles.length; x++) {
      if (profiles[x].getAttribute("id") == name) {
        val = profiles[x];
        break;
      }
    }
    return val;
  },
  
  loadProfile: function(name)
  {
    if (!this.mImporting) {
      this.saveCurrent();
      this.mObsService.notifyObservers(null, "forecastfox:profile:start", "");
    }
    var profile = this.getProfileNode(name);

    if (!profile) {
      //profile doesn't exist - error
      this.mObsService.notifyObservers(null, "forecastfox:profile:error", "Profile not found");
      return;
    }

    var prefs = profile.getElementsByTagName("pref");
    var pref, type, value;

    for (var x = 0; x < prefs.length; x++)
      this.loadPref(prefs[x]);
    
    this.mProfile = name;
    
    if (!this.mImporting)
      this.mObsService.notifyObservers(null, "forecastfox:profile:end", name);
  },
  
  loadPref: function(prefnode)
  {
    var pref  = prefnode.getAttribute("name");
    var type  = prefnode.getAttribute("type");
    var value = prefnode.getAttribute("value");
    
    switch (type) {
      case "Bool":
        value = (value == "true") ? true : false;
        break;
      case "Int":
        value = Number(value);
        break;
      case "Char":
        value = "'" + value + "'";
    }

    //now set the pref
    eval("this.mPrefs.set" + type + "Pref('" + pref + "', " + value + ")");
  },

  copyCacheFiles: function( p1, profile_2 )
  {
    // copy cache files and prefs of p1 to p2 name - p1 is a string - profile_2 is a profile node
    var p2 = profile_2.getAttribute("id");
    
    // first rename the files
    this.mFileIO.copyFile("currentcache-" + p1 + ".xml"  , "currentcache-" + p2 + ".xml");
    this.mFileIO.copyFile("forecastcache-" + p1 + ".xml" , "forecastcache-" + p2 + ".xml");
    this.mFileIO.copyFile("linkscache-" + p1 + ".xml"    , "linkscache-" + p2 + ".xml");

    // then redo the prefs
    var prefs     = [PREF_LINKS_CACHE, PREF_CURRENT_CACHE, PREF_FORECAST_CACHE];
    var isCurrent = (this.mProfile == p2);
    var value     = null;
    
    for (var x = 0; x < prefs.length; x++) {
      value = (isCurrent) ? this.mPrefs.getCharPref(prefs[x]) : this.getProfilePref(p2, prefs[x]);
      value = value.replace(/-(.+?)\.xml$/, "-" + p2 + ".xml");
      
      if (isCurrent)
        this.mPrefs.setCharPref(prefs[x], value);
      
      this.setProfilePref(profile_2, prefs[x], value)
    }
  },
  
  renameProfile: function(old_name, new_name)
  {
    // Filter bad characters out, then change the profile node name
    new_name = this.stripName(new_name);
    this.getProfileNode(old_name).setAttribute("id", new_name);
    
    // Change the profile pref if we renamed the current profile.
    if (this.mPrefs.getCharPref("profile.current") == old_name) {
      this.mProfile = new_name;
      this.mPrefs.setCharPref("profile.current", new_name);
    }

    // copy cache files to new name
    this.copyCacheFiles(old_name, this.getProfileNode(new_name));
    this.saveProfiles();

    this.mObsService.notifyObservers(null, "forecastfox:profile:rename", new_name);    
    return new_name;
  },

  stripName: function(name)
  {
    // remove illegal characters
    return name.replace(/[\\\/:*?<>|\'\"]/g, "")
  },
  
  deleteAllProfiles: function()
  {
    var profiles = this.mDoc.getElementsByTagName("profile");
    while (profiles.length > 0)
      this.deleteProfile(profiles[0].getAttribute("id"));
  }, 
  
  deleteProfile: function(name)
  {
    var node = this.getProfileNode(name);
    
    // remove cached files
    this.mFileIO.removeFile("currentcache-" + name + ".xml");
    this.mFileIO.removeFile("forecastcache-" + name + ".xml");
    this.mFileIO.removeFile("linkscache-" + name + ".xml");
    
    //remove the node
    node.parentNode.removeChild(node);
    
    //save it
    this.saveProfiles();
    
    this.mObsService.notifyObservers(null, "forecastfox:profile:remove", name);
  },

  getProfiles: function()
  {
    var text = this.mFileIO.read(PROFILE_FILE);

    if (!text)
      return this.migrateProfiles();
  
    var doc = this.mDOMParser.parseFromString(text, "text/xml");
    
    if (doc.firstChild.nodeName == "parsererror")
      return this.migrateProfiles();
      
    return doc;
  },
  
  saveProfiles: function()
  {
    //save the doc to file
    this.mFileIO.write(PROFILE_FILE, this.mSerializer.serializeToString(this.mDoc));
  },
  
  filePicker: function(type)
  {
    var sbs           = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
    var bundle        = sbs.createBundle("chrome://forecastfox/locale/forecastfox.properties");
    var nsIFilePicker = Components.interfaces.nsIFilePicker;
    var picker        = Components.classes["@mozilla.org/filepicker;1"].createInstance(nsIFilePicker);
    var windowManager = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService().QueryInterface(Components.interfaces.nsIWindowMediator);
    var topWindow     = windowManager.getMostRecentWindow(null);
    
    switch (type) {
      case "export":
        picker.init(topWindow, bundle.GetStringFromName("ff.profiles.export"), nsIFilePicker.modeSave);
        picker.appendFilters(nsIFilePicker.filterXML);
        picker.defaultExtension = ".xml";
        break;
      case "import":
        picker.init(topWindow, bundle.GetStringFromName("ff.profiles.import"), nsIFilePicker.modeOpen);
        picker.appendFilters(nsIFilePicker.filterXML);
        picker.defaultExtension = ".xml";
        break;
    }
    
    // get the file and its contents
    var res = picker.show();
    if ((res == nsIFilePicker.returnOK) || (res == nsIFilePicker.returnReplace))
      return picker.file;
    else if (res == nsIFilePicker.returnCancel)
      return false;
  },
  
  exportSettings: function()
  {
    var file = this.filePicker("export");
    if (file == false)
      return;
      
    var alerts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    var sbs    = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
    var bundle = sbs.createBundle("chrome://forecastfox/locale/forecastfox.properties"); 
    var windowManager = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService().QueryInterface(Components.interfaces.nsIWindowMediator);
    var topWindow     = windowManager.getMostRecentWindow(null);
    
    var prefs = EXPORT_ADDITIONAL_PREFS.concat(this.mPrefList);
    
    var profiles = this.mDoc.getElementsByTagName("profiles")[0];
    var doc = this.getExportDoc(EXPORT_FILE);
         
    var importnode = doc.getElementsByTagName("forecastfox-settings")[0];
    importnode.setAttribute("version", VERSION);
    importnode.setAttribute("id", EXPORT_FILE_ID);
    importnode.appendChild(profiles.cloneNode(true));
    
    var current = doc.createElement("current");
    current.setAttribute("id", "settings:current");
    var pref = null;
    

    for (var x = 0; x < prefs.length; x++) {
      pref = this.mDoc.createElement("pref");
      pref.setAttribute("name", prefs[x][0]);
      pref.setAttribute("type", prefs[x][1]);
      pref.setAttribute("value", eval("this.mPrefs.get" + prefs[x][1] + "Pref('" + prefs[x][0] + "')"));
      current.appendChild(pref);
    }
    
    importnode.appendChild(current);
    
    var contents = this.mSerializer.serializeToString(doc);
    
    if (!file.exists())
      file.create(file.NORMAL_FILE_TYPE, 0664);
    
    //make sure file is writeable
    if (!file.isWritable()) 
      file.permissions = 0644;

    var flags = 0x02 | 0x08 | 0x20;
    var mFoStream = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
    mFoStream.init(file, flags, 0664, 0);
    mFoStream.write(contents, contents.length);
    mFoStream.close();
    
    alerts.alert(topWindow, bundle.GetStringFromName("ff.alert"), bundle.GetStringFromName("ff.profiles.export.success"));
  },

  importSettings: function()
  {
    var file = this.filePicker("import");
    if (file == false)
      return;
      
    var alerts   = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    var sbs      = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
    var bundle   = sbs.createBundle("chrome://forecastfox/locale/forecastfox.properties");
    var windowManager = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService().QueryInterface(Components.interfaces.nsIWindowMediator);
    var topWindow     = windowManager.getMostRecentWindow(null);
    var doc      = this.mDOMParser.parseFromString(this.mFileIO.readFile(file), "text/xml");
    try {
      // mozilla doesn't have a version-checker
      var vchecker = Components.classes["@mozilla.org/updates/version-checker;1"].getService(Components.interfaces.nsIVersionChecker);
    } catch (e) { vchecker = false; }
    
    try {
      // check versions and validity
      var file_version = doc.getElementsByTagName("forecastfox-settings")[0].getAttribute("version");
      
      if (vchecker &&
            (vchecker.compare(file_version, LAST_COMPATIBLE_VERSION) < 0 ||
             vchecker.compare(file_version, VERSION) > 0))
          throw "incomp";
   
      if (!this.validateFile(doc))
        throw "invalid";
        
      // deconstruct the settings file and load the settings
      var profs_node = doc.getElementsByTagName("profiles")[0];
      var current    = doc.getElementsByTagName("current")[0];
      
      var prefnodes  = current.getElementsByTagName("pref");

      this.mImporting = true;
      this.mObsService.notifyObservers(null, "forecastfox:profile:start", "");

      // now replace profiles with the ones in the import file
      this.deleteAllProfiles();
      
      var profiles = profs_node.getElementsByTagName("profile");
      var profilesnode = this.mDoc.getElementsByTagName("profiles")[0];

      for (var x = profiles.length - 1; x >= 0; x--) {
        this.mObsService.notifyObservers(null, "forecastfox:profile:new", profiles[x].getAttribute("id"));
        profilesnode.appendChild(profiles[x]);
      }

      for (var x = 0; x < prefnodes.length; x++)
          this.loadPref(prefnodes[x]);

      this.mObsService.notifyObservers(null, "forecastfox:profile:end", this.mPrefs.getCharPref("profile.current"));
      this.mImporting = false;
    }
    
    catch (e) {
      alerts.alert(topWindow, bundle.GetStringFromName("ff.alert"), bundle.GetStringFromName("ff.profiles.import." + (e == "incomp" ? "incomp" : "invalid")));
      this.mImporting = false;
      return;
    }
    
    alerts.alert(topWindow, bundle.GetStringFromName("ff.alert"), bundle.GetStringFromName("ff.profiles.import.success"));
  },
  
  ///////////////////////////
  // file validation
  
  validateFile: function(doc)
  {
    // make sure the file is well-formed and has all the parts it needs
    if ((doc.getElementsByTagName("profiles").length < 1) || (doc.getElementsByTagName("current") < 1) ||
        (doc.getElementsByTagName("forecastfox-settings")[0].getAttribute("id") != EXPORT_FILE_ID))
      return false;

    // make sure all the prefs are forecastfox prefs - don't want other prefs being set
    var profiles = doc.getElementsByTagName("profile");
    for (var x = 0; x < profiles.length; x++)
      if (!this.validateProfile(profiles[x]))
        return false;

    if (!this.validateProfile(doc.getElementsByTagName("current")[0]))
      return false;

    return true;
  },
  
  validateProfile: function(profile_node)
  {
    var prefs = profile_node.getElementsByTagName("pref");
    for (var x = 0; x < prefs.length; x++)
      if (!this.validatePref(prefs[x]))
        return false;
    
    return true;
  },
  
  validatePref: function(pref_node)
  {
    var valid_prefs = EXPORT_ADDITIONAL_PREFS.concat(this.mPrefList);

    for (var x = 0; x < valid_prefs.length; x++)
      if (valid_prefs[x][0] == pref_node.getAttribute("name"))
        return true;
        
    return false;
  },
  
  ///////////////////////////
  // utility functions for moving files from defaults directory
  
  migrateProfiles: function()
  {
    // In this function, we copy the profiles.xml file from the defaults folder to the profile directory
    var target = this.mDirService.get("ProfD", Components.interfaces.nsIFile);
    target.append("forecastfox");
    
    // Delete the forecastfox directory if its there
    if (target.exists())
      target.remove(true);
      
    target.create(target.DIRECTORY_TYPE, 0755);

    var file = this.getDefaultFile(PROFILE_FILE);
    
    file.copyTo(target, PROFILE_FILE);
    
    target.append(PROFILE_FILE);
    if (!target.isWritable()) 
      target.permissions = 0644;
      
    this.mPrefs.setCharPref("profile.current", "default");
    
    return this.getProfiles();
  },
  
  getExportDoc: function()
  {
    return this.mDOMParser.parseFromString(this.mFileIO.readFile(this.getDefaultFile(EXPORT_FILE)), "text/xml");
  },
  
  getDefaultFile: function(file_name)
  {
    var file = this.mDirService.get("ProfD", Components.interfaces.nsIFile);
    file.append("extensions");
    file.append(GUID);
    file.append("defaults");
    file.append(file_name);
   
    if (!file.exists()) {
      file = this.mDirService.get("XCurProcD", Components.interfaces.nsIFile); 
      file.append("extensions");
      file.append(GUID);
      file.append("defaults");
      file.append(file_name);
    }
   
    return file;
  },
  
  ///////////////////////////
  // nsITimerCallback
  notify: function(timer)
  {
    if (timer == this.mTimer)
      this.rotateProfile();
  },
  
  /////////////////////////////
  // nsIObserver
  observe: function(aSubject, aTopic, aData)
  {
    switch (aTopic) {
      case "nsPref:changed":
        var new_prof = this.mPrefs.getCharPref("profile.current");
        if ((aData == "profile.current") && (new_prof != this.mProfile)) {
          this.loadProfile(this.mPrefs.getCharPref("profile.current"));
          this.scheduleRotate();
        }
        else if ((aData == "profile.switch.delay") || (aData == "profile.switch.enabled"))
          this.scheduleRotate();
        break;
      case "xpcom-shutdown":
        //save the current profile
        this.saveCurrent();

        this.mObsService.removeObserver(this, "xpcom-shutdown");
        var pbi = this.mPrefs.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
        pbi.removeObserver("", this);

        // Release strongly held services.
        this.mPrefs     = null;
        this.mObsService  = null;
        this.mDOMParser   = null;
        this.mSerializer  = null;

        this.mFileIO.destroy();
        this.mFileIO      = null;
        break;
    }
  },

  QueryInterface: function (iid)
  {
    if (!iid.equals(Components.interfaces.nsIForecastfoxProfiles) &&
        !iid.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
};

/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/

var gModule =
{
    mFirstTime: true,
    registerSelf: function (aCompMgr, aFileSpec, aLocation, aType)
    {
    if (this.mFirstTime) {
      this.mFirstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    for (var key in this.mObjects) {
      var obj = this.mObjects[key];
      aCompMgr.registerFactoryLocation(obj.CID, obj.className, obj.contractID,
                                       aFileSpec, aLocation, aType);
    }
  },

  getClassObject: function (aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    for (var key in this.mObjects) {
      if (aCID.equals(this.mObjects[key].CID))
        return this.mObjects[key].factory;
    }

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  mObjects: {
    updateService: {
      CID: Components.ID(GUID),
      contractID: "@mozdev.org/forecastfox/update;1",
      className: "ForecastFox Update Service",
      factory: {
        createInstance: function (aOuter, aIID)
        {
          if (aOuter != null)
            throw Components.results.NS_ERROR_NO_AGGREGATION;
          return (new nsForecastfoxUpdate()).QueryInterface(aIID);
        }
      }
    },

    profileService: {
      CID: Components.ID("{39C873BF-E30C-4df2-A3DB-4A59AE001EC9}"),
      contractID: "@mozdev.org/forecastfox/profiles;1",
      className: "ForecastFox Profiles Service",
      factory: {
        createInstance: function (aOuter, aIID)
        {
          if (aOuter != null)
            throw Components.results.NS_ERROR_NO_AGGREGATION;
          return (new nsForecastfoxProfiles()).QueryInterface(aIID);
        }
      }
    }
  },

  canUnload: function(compMgr)
  {
      return true;
  }
};

function NSGetModule(compMgr, fileSpec) { return gModule; }

/******************************************************************************
 * Utility functions
 ******************************************************************************/
function addUnit(base, unit, link) {
  if ((base == "N/A") || ((!Number(base)) && (base != 0)))
    return base;
  else
    return base + (link ? link : "") + unit;
}

function fileIO()
{
  this.mDirService    = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);
  this.mFoStream    = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
  this.mFiStream    = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
  this.mSiStream    = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);

  return this;
}

fileIO.prototype = {
  destroy: function()
  {
    this.mDirService  = null;
    this.mFoStream    = null;
    this.mFiStream    = null;
    this.mSiStream    = null;
  },
  
  write: function(name, contents)
  {
    var file = this.mDirService.get("ProfD", Components.interfaces.nsIFile);

    //put in forecastfox subdirectory
    file.append("forecastfox");
    if (!file.exists())
      file.create(file.DIRECTORY_TYPE, 0755);

    //check if file exists and isn't a directory
    file.append(name);
    if (!file.exists())
      file.create(file.NORMAL_FILE_TYPE, 0664);
    else {
      if (file.isDirectory()) {
        file.remove(true);
        file.create(file.NORMAL_FILE_TYPE, 0664);
      }
    }
    
    //make sure file is writeable
    if (!file.isWritable()) 
      file.permissions = 0644;
            
    var flags = 0x02 | 0x08 | 0x20;
    this.mFoStream.init(file, flags, 0664, 0);
    this.mFoStream.write(contents, contents.length);
    this.mFoStream.close();

    return true;
  },
  
  readFile: function(file)
  {
    if (!file.exists())
      return false;
      
    var contents = new String();
    try {
      this.mFiStream.init(file, 1, 0, false);
      this.mSiStream.init(this.mFiStream);
      var available = this.mSiStream.available();
      contents += this.mSiStream.read(available);

      this.mSiStream.close();
      this.mFiStream.close();

      return contents;
    }
    catch(e) {
      return false;
    }
  },

  read: function(name)
  {
    var file = this.mDirService.get("ProfD", Components.interfaces.nsIFile)

    //get from forecastfox subdirectory
    file.append("forecastfox");
    if (!file.exists())
      return false;

    //check if file exists and isn't a directory
    file.append(name);
    
    if (!file.exists())
      return false;
      
    if (file.isDirectory()) {
      file.remove(true);
      return false;
    }
    
    return this.readFile(file);
  },
  
  copyFile: function(file, newfile)
  {
    var old_file = this.mDirService.get("ProfD", Components.interfaces.nsIFile);
    old_file.append("forecastfox");
    
    // remove the file if it already exists
    var file_check = old_file.clone();
    file_check.append(newfile);
    if (file_check.exists())
      file_check.remove(false);

    old_file.append(file);
    if (!old_file.exists())
      return false;
    
    // copy the old one to this location
    old_file.copyTo(null, newfile);
    
    return true;
  },
  
  removeFile: function(file)
  {
    var target = this.mDirService.get("ProfD", Components.interfaces.nsIFile);
    target.append("forecastfox");
    target.append(file);
    try {
      target.remove(false);
    } catch (e) { }
    
  }
  
};

function getWindow()
{ 
  var windowManager = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService();
  var windowManagerInterface = windowManager.QueryInterface( Components.interfaces.nsIWindowMediator); 
  var topWindow = windowManagerInterface.getMostRecentWindow("navigator:browser");
  if (!topWindow)
    topWindow = window.openDialog("chrome://browser/content/browser.xul", "_blank", "chrome,all,dialog=no", "about:blank", null, null);
  return topWindow;
};
